//-----------------------------LICENSE NOTICE------------------------------------
//
////////////////////////////////////////////////////////////////////////////////
///////////////////////////  SHOOT AND RUN /////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
//
//  Code & Graphics Copyright (C) 2020 Rafa Gutierrez (rafagutierrez@qq.com)
//
//  This file is part of Shoot and run: an Amstrad CPC Game made with CPCTelera
//  for CPCRetroDev 2020
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.


#include "entity.h"

//==============================================================================
// Note 1: 
// We do not control the max number of entities we can create
// For a 8bit machine game better not to control things are inside the range
// otherwise we will waste too much code, space and time
// so we just guarantee that the pre/post conditions of the functions
//
// Note 2:
// If we force the compiler to operate with temporal variables, 
// the compiler will optimize the code and try to use the registers!!!
// If we use global variables the compiler optimizations will be limited
// TIP: Calculate with temporals and save the result in global
//==============================================================================

//==============================================================================
// PRIVATE DATA MEMBERS
//==============================================================================
// do not initialize global variables (SDCC issue RAM/ROM)
//
//
// HACK: Since compiler will allocate these variable together, we add a byte 
// after the m_entities array that will be initialized to e_type_invalid to 
// guarantee there is always a stop condition for the loop
// In this way we don't need to declare the size of m_entities as MAX_ENTITES+1
Entity_t  m_entities[MAX_ENTITIES];        // Entities
u8 m_zero_type_at_the_end;                 // A zero-byte to signal invalid entity at the end

// Note1: if we define m_zero_type_at_the_end as a constant the compiler doesn't put them
// together on memory!
//
// Note2: In a 8bit machine, an alternative to be compatible with any compiler: 
// m_entities and m_zero_type_at_the_end inside a struct.
//
// struct {
//    Entity_t  m_entities[MAX_ENTITIES];        // Entities
//    u8 m_zero_type_at_the_end;                 // A zero-byte to signal invalid entity at the end
// } Storage_t;
// Storage_t storage;
//
// Note 3: If machine is not 8bit careful with the padding, depends the size used for the variables
// it might add some padding between the fields of the structure

Entity_t* m_next_free_entity;              // Pointer to next free entity
u8 m_num_entities;                         // Number of entities

//==============================================================================
// PUBLIC FUNCTIONS
//==============================================================================

////////////////////////////////////////////////////////////////////////////////
// INIT
//      Initializes the entity manager. Zeroes everything
//
void man_entity_init() {
    cpct_memset (m_entities, 0, sizeof(m_entities));    // sizeof(m_entities) = size of all the array
    m_next_free_entity = m_entities;                    // same as m_next_free_entity = &m_entities[0]
    m_num_entities = 0;
    m_zero_type_at_the_end = e_type_invalid;            
}


////////////////////////////////////////////////////////////////////////////////
// CREATE
//      Creates a new entity and returns a pointer to it
// Returns:
//      Entity_t*       Pointer to the new created entity
// Preconditions:
//      There must be space available in the m_entities array.
//      Otherwise, it will assign memory beyond the array.
//
Entity_t* man_entity_create() {
   Entity_t* e = m_next_free_entity;
   //++m_next_free_entity;                          // Compiler does a bad work with optimization
   m_next_free_entity = e + 1;                      // Compiler optimize since we use temp to calculate
   e->type = e_type_default;
   e->cmps = e_cmp_default;
   ++m_num_entities;
   return e;
}

///////////////////////////////////////////////////////////////////////////////
// CLONE
//      Creates a clone from an entity
// Returns:
//      Entity_t*       Pointer to the new clone entity
// Preconditions:
//      There must be free space for at least one more entity
//
Entity_t* man_entity_clone(Entity_t* e) {
    Entity_t* clone = man_entity_create();
    cpct_memcpy(clone, e, sizeof(Entity_t));
    return clone;
}


///////////////////////////////////////////////////////////////////////////////
// RESTORE TEMPLATE
//      Restore
// 
// Input: 
//    0 e:       Pointer to the entity to be restored
//    - templ_e: Pointer to the constant template entity
//
void man_entity_restore_template(Entity_t* e, const Entity_t* templ_e) {
    cpct_memcpy(e, templ_e, sizeof(Entity_t));
}


////////////////////////////////////////////////////////////////////////////////
// CREATE TEMPLATE ENTITY
//    Creates a new entity and instantiates it using values from
// a constant template entity.
// Input:
//    - templ_e: Pointer to the constant template entity
// Return:
//    - e: New entity created and instantiated
//
Entity_t* man_entity_create_from_template(const Entity_t* templ_e) {
   Entity_t* e = man_entity_create();                
   cpct_memcpy(e, templ_e, sizeof(Entity_t));        
   return e;
}



///////////////////////////////////////////////////////////////////////////////
// NOTE: The parameter is a pointer to a function
// A pointer to a function is just a memory address but the compiler needs to know
// the type of the pointer so when it generates the code for that function call 
// it also generate the code to deal with the parameters and the returned type of the function
//
// Note: the compiler translates a function call as:
//      call to a label + code to add function parameters and return value on the stack
//
// So a pointer to a function must know what are the parameters and the type of the return value
//
// Syntaxis is the same as the function declaration BUT add (* ) to the name of the function
// to indicate that it's a pointer to that function
//
// void (*ptrfunc)(Entity_t*)  
// ptrfunc is a pointer to function
// returns: void 
// parameters: pointer to Entity_t
//
// To loop all the entities
// Option A: Allow invalid entities among valid entities
//           [E1][-][E3][-][E7][-][-] + number of valid entities 
//           To delete, just change the type to invalid
//           CONS: To loop all or insert,  check type and number of valid entities still to process
//
// Option B: Do not allow invalid entities among valid entities
//           [E1][E2][E3][E7][-][-]
//           To delete E2, option 1. E3->E2 and E7->E3 (slower but keeps the order)
//           To delete E2, option 2. E7->E2 and invalidate E7 (faster but changes the order)
//           CONS: Change the place of entities in memory 
//           After delete E2 any previous pointer to E7 would have the wrong address

///////////////////////////////////////////////////////////////////////////////
// FORALL
// Calls a given update function for all valid entities. Update
// function is given as parameter and receives a pointer to an
// entity and returns void.
//
// Input:
//  updfunc: Pointer to the update function
void man_entity_forall( UpdateFunc_t updFunc ) {
     // Use a local variable instead of manipulate a global variable (better processor optimization)
    Entity_t* e = m_entities;           // e points to m_entities (entities array first entity)

    // Option B2 (no invalid entities among valid entities)
    // loop precondition: there is always at least one invalid entity 
    while ( IsValid(e) ) {
        updFunc(e);                  // calls the function received as parameter
        ++e;
    }
}


///////////////////////////////////////////////////////////////////////////////
// FORALL MATCHING
// Loops through all the valid entities but only calls the function if entity matches signature
//
// Input:
//   updfunc: Pointer to the update function
//
// Potential upgrade: avoid all systems loop through all entities
//
void man_entity_forall_matching( UpdateFunc_t updFunc, u8 signature ) {
     // Use a local variable instead of manipulate a global variable (better processor optimization)
    Entity_t* e = m_entities;           // e points to m_entities (entities array first entity)

    // Option B2 (no invalid entities among valid entities)
    // loop precondition: there is always at least one invalid entity 
    while ( IsValid(e) ) {
        if ( CmpsSignatureMatches(e, signature) )       // check if entity matches the signature
        updFunc(e);                                     // calls the function received as parameter
        ++e;
    }
}


///////////////////////////////////////////////////////////////////////////////
// FORALL PAIRS MATCHING
// Calls an updateFunction for all entities that match a given signature.
//
// Input:
//   updfunc:   Pointer to the update function void(*)(Entity_t*, Entity_t*)
//   signature: Bitset with the components the entity must have (all of them)
//
void man_entity_forall_pairs_matching( UpdateFunc2E_t updFunc, u8 signature ) {
     // Use a local variable instead of manipulate a global variable (better processor optimization)
    Entity_t* e_left = m_entities;                           // points to array 1st entity 

    // Option B2 (no invalid entities among valid entities)
    // loop precondition: there is always at least one invalid entity 
    while ( IsValid(e_left) ) {
        if ( CmpsSignatureMatches(e_left, signature) ) {         // check entity signature
            Entity_t* e_right = e_left + 1;                  // e_right initial value
            while ( IsValid(e_right) ) {
                if ( CmpsSignatureMatches(e_right, signature) )  // check entity signature
                    updFunc(e_left, e_right);                // calls function received as parameter
                ++e_right;
            }
        }
        ++e_left;
    }
}


//////////////////////////////////////////////////////////////////////////////
// DESTROY
//      Destroys a given entity and frees up its assigned memory
//
// Input:
//      dead_e: Pointer to the entity to be destroyed
// Preconditions:
//      dead_e must point to a valid entity
//
// Copy last entity to the entity we want to delete and decrease pointer m_next_free_entity
// If we want to delete the last entity then no need to copy, just set entity to invalid entity
void man_entity_destroy(Entity_t* dead_e) {
    Entity_t* de = dead_e;                              // to optimize the code generated by compiler
    //Entity_t* last = m_next_free_entity-1;            // do not operate with globals (bad optimization)
    Entity_t* last = m_next_free_entity;
    --last;
    if ( de != last)                                    // if dead_e = last no need to copy
        cpct_memcpy(dead_e, last, sizeof(Entity_t));
    last->type = e_type_invalid;
    m_next_free_entity = last;
    --m_num_entities;
}

//////////////////////////////////////////////////////////////////////////////
// SET 4 DESTRUCTION
//      Marks an entity for later destruction
//
// Input:
//      dead_e: Pointer to the entity to be marked
// Preconditions:
//      dead_e must point to a valid entity
//
// OR with a mask to set the destruction bit
// Note: It activates the render component and disactivate the other components
void man_entity_set4destruction(Entity_t* dead_e) {
    dead_e->type |= e_type_dead;                     // dead_e->type = dead_e->type | e_type_dead;          
    dead_e->cmps =  e_cmp_render;                    // Guarantees dead_entity does not update physics, ia, etc.
}


/////////////////////////////////////////////////////////////////////////////
// UPDATE
//      Updates Entity Manager by destroying all marked entities
//
void man_entity_update() {
    Entity_t* e = m_entities;

    while ( IsValid(e) ) {
        if ( IsDead(e) ) {                           //  bit mask with 0x80 to check upper bit
           man_entity_destroy(e);
        }
        else {
           ++e;                                      //  Only increments pointer if no destruction
        }
    }
}

/////////////////////////////////////////////////////////////////////////////
// FREE SPACE
//      Returns the number of free entities available
// Return:
//      u8: Number of free entities
//
u8 man_entity_freeSpace() {
    return MAX_ENTITIES - m_num_entities;
}