//-----------------------------LICENSE NOTICE------------------------------------
//
////////////////////////////////////////////////////////////////////////////////
///////////////////////////  SHOOT AND RUN /////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
//
//  Code & Graphics Copyright (C) 2020 Rafa Gutierrez (rafagutierrez@qq.com)
//
//  This file is part of Shoot and run: an Amstrad CPC Game made with CPCTelera
//  for CPCRetroDev 2020
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.


#pragma once                    // to avoid including several times
#include <cpctelera.h>
#include <man/animations.h>


//==============================================================================
// PUBLIC CONSTANTS
//==============================================================================

//------------------------------------------------------------------------------
// Entity_t Types (used as bit masks)
#define e_type_invalid             0x00    // Invalid entity type. Entity does not exist
#define e_type_player              0x01    // 
#define e_type_enemy               0x02    //
#define e_type_player_vshot        0x04    //
#define e_type_player_hshot        0x08    //
#define e_type_enemy_shot          0x10    //
#define e_type_enemy_bot           0x20    //
#define e_type_empty               0x40    //
#define e_type_dead                0x80    // Upper bit signals dead entity b10000000
#define e_type_default     e_type_empty   

// Entity_t Components (used as bit masks)
#define e_cmp_render               0x01    // Renderable entity
#define e_cmp_movable              0x02    // Movable  entity
#define e_cmp_input                0x04    // Entity controllable by input
#define e_cmp_ai                   0x08    // Entity controllable by artificial intelligence
#define e_cmp_animated             0x10    // Animated entity
#define e_cmp_collider             0x20    // Entity that can be collided by other entities
#define e_cmp_default              0x00    // Default components (no components) 

// Macros (to make code easier to read)
#define IsValid(E)                 ((E)->type != e_type_invalid)
#define IsDead(E)                  ((E)->type & e_type_dead)
#define CmpsSignatureMatches(E, S) (((E)->cmps & (S)) == (S))
#define TypeMatches(E, S)          (((E)->type & (S)) == (S))


//==============================================================================
// PUBLIC DATA STRUCTURES
//==============================================================================

//------------------------------------------------------------------------------
// Maximum number of entities
#define MAX_ENTITIES    32

//------------------------------------------------------------------------------
// Forward declaration of Entity_t type
typedef struct Entity_t Entity_t;               // alias = Entity_t (forward declaration)

//------------------------------------------------------------------------------
// Function pointer types
typedef void (*BehaviourFunc_t)(Entity_t*);           // alias = BehaviourFunc_t (pointer to function)
typedef void (*UpdateFunc_t)(Entity_t*);              // alias = UpdateFunc_t    (pointer to function)
typedef void (*UpdateFunc2E_t)(Entity_t*, Entity_t*); // alias = UpdateFunc2E_t  (pointer to function)  

//------------------------------------------------------------------------------
// Entity_t structure
struct Entity_t {
   u8                   type;
   u8                   cmps;                // components
   u8                   x, y;
   u8                   w, h;   
   i8                   vx, vy;
   u8 const *           sprite;
   BehaviourFunc_t      ai_behaviour;       	// same as: void (*ai_behaviour)(Entity_t*);
   u8                   ai_counter;
   AnimFrame_t const *  anim;                // anim pointer can change, but not the value pointed by anim 
   u8                   anim_counter;     
   u8                   anim_status;         // anim status
   u8                   collides_against;    // against what types of entities might collide                    
};
// Note: 
// We want to be able to change which animation is performed by an entity, for example when 
// an enemy explodes we will change the pointer of the animation to a different one BUT the
// struct AnimFrame_t of any animation cannot be changed

// debug only
extern u8 m_num_entities;                         // Number of entities

//==============================================================================
// PUBLIC INTERFACE
//==============================================================================
void            man_entity_init();
Entity_t*       man_entity_create();
Entity_t*       man_entity_clone(Entity_t* e);
Entity_t*       man_entity_create_from_template(const Entity_t* templ_e);
void            man_entity_restore_template(Entity_t* e, const Entity_t* templ_e);
void            man_entity_destroy(Entity_t* dead_e);
void            man_entity_forall( UpdateFunc_t updFunc );
void            man_entity_forall_matching( UpdateFunc_t updFunc, u8 signature );
void            man_entity_forall_pairs_matching( UpdateFunc2E_t updFunc, u8 signature );
void            man_entity_set4destruction(Entity_t* dead_e);
void            man_entity_update();
u8              man_entity_freeSpace();


//------------------------------------------------------------------------------
// Note 1:
// Forward declaration: to be able to use the structure we define inside its definition
// We need to first declare Entity_t as a struct: struct Entity_t;
// Then we can use Entity_t inside the definition of the struct Entity_t
//
// Note 2:
// Type definition syntax:
// For XXX type 
// 			 typedef XXX alias
// For pointer to function:
// 			 typedef return_type (*alias)(function_parameters_type)
//
// Pointers to functions need to point to a function with same return/parameters type as its declaration
// Exemple:
// void (*ai_behaviour)(Entity_t*);
// ai_behaviour is a pointer to a function that returns void and has a pointer to Entity_t as parameter