//-----------------------------LICENSE NOTICE------------------------------------
//
////////////////////////////////////////////////////////////////////////////////
///////////////////////////  SHOOT AND RUN /////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
//
//  Code & Graphics Copyright (C) 2020 Rafa Gutierrez (rafagutierrez@qq.com)
//
//  This file is part of Shoot and run: an Amstrad CPC Game made with CPCTelera
//  for CPCRetroDev 2020
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.


#include "game.h"
#include <man/entity.h>
#include <man/animations.h>
#include <man/scoreboard.h>
#include <man/livesboard.h>
#include <man/level.h>
#include <sys/input.h>
#include <sys/physics.h>
#include <sys/render.h>
#include <sys/ai.h>
#include <sys/animations.h>
#include <sys/collision.h>          
#include "entity_templates.h"
#include <man/custom_sprites.h>

#include <sprites/new_player.h>
#include <sprites/vshot.h>
#include <sprites/hshot.h>
#include <sprites/star_bot.h>
#include <sprites/empty.h>


//==============================================================================
//==============================================================================
// PRIVATE DATA MEMBERS
//==============================================================================
//==============================================================================
// Global variables cannot be initialized in their declaration 
u8              m_level;                    // Current level              
u8              m_player_x;                 // Player x position
u8              m_lane0_enemy_x;            // Shooter x position
u8              m_star;                     // 0: no star 2: star on right of player 1: star on left of player 
u8              m_player_vshot;             // Player vshot flag   
u8              m_player_lhshot;            // Player left hshot flag   
u8              m_player_rhshot;            // Player right hshot flag   
u8              m_enemy_shot;               // Enemy shot flag   
u8              m_enemy_bot;                // Enemy  bot  flag
u8              m_enemy_bounces;            // Enemies bouncing flag  (does not include mothership)
u8              m_enemy_hides;              // Enemies temporary hide flag
u8              m_bar_lvl;                  // Player weapon temperature level
u8              m_lane_status[3];           // 00: empty, 01: one enemy, 10: two minions, 11: one minion
u8              m_lane2_enemies2drop;       // Number of enemies that need to drop from lane2
u8              m_lane1_enemies2drop;       // Number of enemies that need to drop from lane1
u8              m_num_hshot;                // Every 4 hshot increase the weapon bar temperature
u8              m_abduction_light;          // Mothership abduction light catched the player
u8              music_on;
u8              m_player_avatar;            // Player is human or alien
u8              m_difficulty;

const Entity_t* m_player_tmpl;              // Points to the right player template
const Entity_t* m_lives_tmpl;               // Points to the right lives template


//==============================================================================
//==============================================================================
// PRIVATE FUNCTIONS
//==============================================================================
//==============================================================================

// In CPC there are 6 interruptions per frame (6*50 frames/second = 300Hz)
// Wait for 2 interruptions guarantees raster is not in the VSYNC zone
// VSYNC flag is active 50 times per second, 50Hz = 50 frames/second = every 20.000ms (19.968ms)
// wait(10) = 5 frames per second

void wait(u8 n) {
   do {
      cpct_waitHalts(2);         // cpu stops until next interruption (2 times)
      cpct_waitVSYNC();          // active wait until VSYNC flag = 1 (lowest part of the screen)   
   } while (--n);

}


// it does not update the collision system
void wait_game_cycles(u8 cycles) {

   m_enemy_hides = 0;                              // Show the enemies in case they were hidden
   hidden_counter = 15;

   for (u8 i = cycles; i > 0; i--) {
         sys_ai_update();
         sys_physics_update();                 
         sys_animation_update();   
         wait(CYCLE_DELAY);
         sys_render_update();
         man_entity_update(); 
   } 

   m_enemy_hides   = (m_level > 7) ? 1 : 0;         // Enemy hidden   activated after level 8

}



////////////////////////////////////////////////////////////////////////////////
// CHECK LANE
//
// Input: 
//    - y coordinate
//
// Return: 
//    - Number of the lane (2, 1 or 0)
//
u8 man_game_check_lane(u8 y) {
   u8 lane = 2;                        
   
   if       ( y > LANE1_Y + 2) lane = 0;                         
   else if  ( y > LANE2_Y + 2) lane = 1;  

   return lane;
}



////////////////////////////////////////////////////////////////////////////////
// MOVES AN ENEMY/MINION TO THE NEXT LANE
//
//  Input: 
//    - e is the entity to move down to the next lane, it should be in a lane > lane 0
//
void man_game_move_next_lane(Entity_t* e) {
   Entity_t* clone = man_entity_clone(e);                              // Create a phantom clone entity to erase trail

   clone->cmps =  e_cmp_render;                                        // Guarantees clone does not update physics, ia, etc.
   man_entity_set4destruction(clone);   
   
   e->y += LANE_DY;


   if (e->y > LANE1_Y + 2) e->ai_behaviour =  man_level_lane0_ai();    // Enemies in lane 0 are able to attack

}



////////////////////////////////////////////////////////////////////////////////
// PLAYER KILLED
//
// It is called AFTER player finishes player_death_animation
//    If player no more lives   -> Destroys entity player
//    If player still has lives -> Resets player and let enemy shoot/deploy bot again
//
void man_game_player_killed(Entity_t* e) {

   wait_game_cycles(15);                                                // Little delay to make it look better :)   
   e->cmps  &= ~e_cmp_animated;                                         // Hide player
   // EMPTY SPRITE
   //e->sprite = spr_player_20;                                           // empty one     
   e->sprite = spr_empty;     

   wait_game_cycles(20);                                                // Little delay to make it look better :)

   man_livesboard_update();                                             // Player loses one live

   if (man_livesboard_check() > 0) {                                    // Player can keep playing
      man_game_player_reset(e);                                         // Do not destroy the player yet
      m_enemy_shot = 0;                                                 // Enemy can shot again after player reset finishes
      m_enemy_bot  = 0;                                                 // Enemy can drop a dron after player reset finishes
   }
   else  
      man_entity_set4destruction(e);
}



////////////////////////////////////////////////////////////////////////////////
// ENEMY KILLED
//
// It is called AFTER enemy finishes enemy_death_animation
//    Free player vshot and lane from enemy e, sets for destruction, updates the 
//    lane status, the enemies still to drop in current lane and increase number 
//    of killed enemies and checks if level finished
//
// Lane status:   0: Empty     1:enemy    2:two minions    3:one minion   
//
void man_game_enemy_killed(Entity_t* e) {   
   u8 lane = man_game_check_lane(e->y);                     
   u8 lane_status = m_lane_status[lane];

   if ((e->type & e_type_enemy_bot) && (m_star != 0)) m_star = 0;                // 0: no star 2: star on right of player 1: star on left of player

   man_entity_set4destruction(e);                                                // Destroy entity

   // Update lane status
   if (m_level < SPLIT_LEVEL)                                                    
      lane_status = 0;                                                           // Free lane from dead enemy
   else                                                                          // Enemies split from first hit
      lane_status = (++lane_status) & 3;                                         // Update to the next lane_status (00, 01, 10, 11)
      
   m_lane_status[lane] = lane_status;                                            

   // Update enemies2drop if necessary
   if ((lane == 2) && (m_lane2_enemies2drop > 0)) m_lane2_enemies2drop--;        
   if ((lane == 1) && (m_lane1_enemies2drop > 0)) m_lane1_enemies2drop--;

   // Update enemies killed in this level
   if (lane_status == 0) ++m_level_enemies_killed;                               // Update kills counter if lane is left empty
   
   if (m_level_last_kill()) man_level_setup(++m_level);                          // Check if level finished

   m_player_vshot = 0;                                                           // Since enemy death animation finished player can vshot again
   
 }


//==============================================================================
//==============================================================================
// PUBLIC FUNCTIONS
//==============================================================================
//==============================================================================

////////////////////////////////////////////////////////////////////////////////
// INIT
//    Sets all values to initial game status
//
void man_game_init() {
// Note: all definitions and declarations should be at the begining of the scope (before any instruction)
// tip: create a scope to define
   man_entity_init();

   // Scoreboard
   man_scoreboard_create();
   man_scoreboard_set(0,0,0,0,0,0);


   // Bar
   man_entity_create_from_template(&bar_tmpl);
   m_bar_lvl = 0;
   
   // No enemies on any lane on start
   cpct_memset(m_lane_status, 0, sizeof(m_lane_status));
   m_lane0_enemy_x = 0;
   m_star = 0;
   
   // No shots on start
   m_player_vshot  = 0;
   m_player_lhshot = 0;
   m_player_rhshot = 0;
   m_num_hshot     = 0;
   m_enemy_shot    = 0;
   m_enemy_bot     = 0;

   // No abductions on start
   m_abduction_light = 0;
   
   // Player
   m_player_x     = PLAYER_X;


   // Player lifes
   if (m_player_avatar == HUMAN) {
      m_player_tmpl = &human_tmpl;
      m_lives_tmpl  = &lives_human_tmpl;
   }
   else {
      m_player_tmpl = &alien_tmpl;
      m_lives_tmpl  = &lives_alien_tmpl;
   }
   
   man_entity_create_from_template(m_player_tmpl);
   man_livesboard_create(m_lives_tmpl);


   // Systems that need initialization
   sys_input_init();
   sys_ai_init();
   sys_render_init(); 


   // Level
   m_level = FIRST_LVL;
   man_level_setup(FIRST_LVL);

}


////////////////////////////////////////////////////////////////////////////////
// PLAY
//    Implements the game main loop
//
void man_game_play() {
   
   while (man_livesboard_check() > 0) {
      sys_ai_update();
      sys_physics_update();
      sys_collition_update();  
      sys_animation_update();   

      wait(CYCLE_DELAY);

      sys_render_update();         
      man_entity_update(); 
   }

}


////////////////////////////////////////////////////////////////////////////////
// RESETS THE GAME
//
// Resets the game
//
//
void  man_game_reset() {

   // Little GAME OVER animation   
   // flash effect
   
   Entity_t* e = man_entity_create_from_template(&flash_tmpl);   
   e->ai_counter = 30;
 
   wait_game_cycles(80);

   wait(50);

   man_entity_forall( man_entity_set4destruction );        // destroy all entities
   wait(1);
   cpct_clearScreen(0);                                    // clean the screen
   man_game_init();                                        // sets all values to initial game status

}



////////////////////////////////////////////////////////////////////////////////
// CREATE ENEMY
//    Creates a new enemy as a son of the Mothership. 
// 
// Precondition: Upper lane (Lane 2) is free (no enemies).
//
// Input:
//    - e_mother: Pointer to the Mothership entity
//
void man_game_create_enemy(Entity_t const* e_mother) {
   // If there is an enemy already on the top lane
   if (m_lane_status[2] != 0) return;        // already checked before call
   
   // Create a new enemy
   {
      Entity_t* e = man_level_create_enemy();  // creates an enemy of the right type for current level
      e->x  = e_mother->x + 4;                 // mothership width = 18, enemy width = 10 center:(18-10)/2
      e->vx = (e_mother->x & 0x07) ? 1 : -1;
   }
   // Mark enemy is on lane
   m_lane_status[2] = 1;                       // lane 2 now has one enemy
}



////////////////////////////////////////////////////////////////////////////////
// ENEMY NEEDS TO DESCEND
//
//      Check if in the current lane there is still any enemy/dron that needs to go down
//
// Input:
//    - current lane : current lane id
//    - enemies2drop : counter of enemies left to drop (in case of minions)
//
// Return:
//     1 if enemy needs to descend
//
// Lane:          Top lane = 2, Middle lane = 1, Bottom lane = 0
// Lane status:   0: Empty     1:enemy    2:two minions    3:one minion    
//   
u8 enemy_needs2descend(u8 current_lane, u8 enemies2drop) {
    u8 needs2descend       = 0;
    u8 lane_below          = current_lane - 1;
    u8 current_lane_status = m_lane_status[current_lane];
    u8 below_lane_status   = m_lane_status[lane_below];

    if (below_lane_status == 0) {                    // lane below is free   
        switch (current_lane_status) {
            case 1:  current_lane_status = 0;        // it's an enemy then lane 1 has an enemy        L2==1 -> L2=0 L1=1
                     below_lane_status   = 1;
                     enemies2drop        = 0;
                     break;
            case 2:  current_lane_status = 3;        // there are two minion then lane 1 has a minion L2==2 -> L2=3 L1=3
                     below_lane_status   = 3;
                     enemies2drop        = 1;
                     break;
            case 3:  current_lane_status = 0;        // it's a minion then lane 1 has a minion        L2==3 -> L2=0 L1=3    
                     below_lane_status   = 3;
                     enemies2drop        = 0;
                     break;
        }  
        needs2descend = 1;
    } 
    else {                                           // lane below is not free but there is a minion to drop
        if (enemies2drop > 0) {                      // only case it is e_enemy is last minion
           current_lane_status = 0;                  // it's last minion then lane 1 has two minion   L2==3 -> L2=0 L1=2    
           below_lane_status   = 2;
           enemies2drop        = 0;
           needs2descend = 1;
        }    
    }   
    m_lane_status[current_lane] = current_lane_status;
    m_lane_status[lane_below]   = below_lane_status;

    if (current_lane == 1) m_lane1_enemies2drop = enemies2drop;
    else                   m_lane2_enemies2drop = enemies2drop;

    return needs2descend;

}   



////////////////////////////////////////////////////////////////////////////////
// ENEMY LANE DOWN
//
// Input:
//    - e_mother: Pointer to enemy entity
//
// Lane:          Top lane = 2, Middle lane = 1, Bottom lane = 0 
//
void man_game_enemy_lane_down(Entity_t* e_enemy) {
   u8 lane = man_game_check_lane(e_enemy->y);  

   // At this point e_enemy is in lane 2 or lane 1
   {  
      u8 enemies2drop = (lane == 2) ? m_lane2_enemies2drop : m_lane1_enemies2drop;
      if (enemy_needs2descend(lane, enemies2drop)) man_game_move_next_lane(e_enemy);
   }
                
}
 

////////////////////////////////////////////////////////////////////////////////
// GAME ENTITY DESTROY
//
// What should happen when an entity of type X dies (besides its destruction)?
// Sets the entity for destruction and add some actions to the types that need it
//
// Input:
//    - e: Entity to destroy
//
void man_game_entity_destroy(Entity_t* e) {   
   // what needs to be done before a type X entity gets destroyed...

   if      ( e->type & e_type_player ) man_game_player_killed(e);
   else if ( e->type & e_type_enemy  ) man_game_enemy_killed(e);
   else { 
         // Shots that did not hit the target but their timer is up   
         if (e->type & e_type_player_vshot)  m_player_vshot = 0;                             // Player can vshot again (vshot timer is up)       
// PROBLEM CAUSE AT THIS POINT e->x might be negative... and even it should be smaller than player is actually bigger
         if ((e->type & e_type_player_hshot) && (m_player_x > e->x)) m_player_lhshot = 0;    // Player can lhshot again (hshot timer is up)

// Change instead of destroy hshot by time do it by x boundaries!


         if ((e->type & e_type_player_hshot) && (m_player_x < e->x)) m_player_rhshot = 0;    // Player can rhshot again (hshot timer is up)
         if (e->type & e_type_enemy_shot)   m_enemy_shot   = 0;                              // Enemy can shot again   (enemy shot timer is up)
         
         man_entity_set4destruction(e);
   }

}


////////////////////////////////////////////////////////////////////////////////
// PLAYER HIT
//
// It is called when player gets hit 
// Flash FX, stop enemies from attacking and sets up the player death animation (not destroy yet)
// When the ai_timer timer is up then man_game_player_killed will check if player can keep playing
//
void man_game_player_hit(Entity_t* e_player) {

   man_entity_create_from_template(&flash_tmpl);                // flash effect 

   m_enemy_bot  = 1;                                            // Stop enemies from keep attacking
   m_enemy_shot = 1;                                            // Stop enemies from keep attacking
   
   e_player->vx = 0;                                            // Stop player
   e_player->cmps &= ~e_cmp_input;                              // Temporary forbid player to control its avatar
   e_player->cmps |=  e_cmp_ai;                                 // Ai takes control
   e_player->ai_behaviour = sys_ai_behaviour_player_dies;
   
}


////////////////////////////////////////////////////////////////////////////////
// ENEMY HIT
//
// It is called when enemy gets hit, sets up the enemy death animation 
// When the ai_timer timer is up then man_game_enemy_killed will free the lane and destroy the enemy entity
//
void man_game_enemy_hit(Entity_t* e) {
   u8 lane = man_game_check_lane(e->y);  
   u8 lane_status = m_lane_status[lane];

   if (lane == 0) m_lane0_enemy_x = 0;                            

   if (m_level < SPLIT_LEVEL) {                             
      e->ai_behaviour = sys_ai_behaviour_enemy_dies;
      return;
   }
       
   // Check lane_status:      0: empty        1: enemy (can split)        2: two minions      3: one minion
    switch (lane_status) {
        case 1:  e->ai_behaviour = sys_ai_behaviour_enemy_splits;
                 break;
        case 2:  e->ai_behaviour = sys_ai_behaviour_enemy_dies;
                 break;
        case 3:  e->ai_behaviour = sys_ai_behaviour_enemy_dies;
    }

}


////////////////////////////////////////////////////////////////////////////////
// BOT HIT
//
// It is called when enemy gets hit 
// Let enemy deploy bot again and destroy the bot entity
//
void man_game_bot_hit(Entity_t* e) {

   m_star = 0;
   m_enemy_bot = 0;
   man_game_entity_destroy(e);

}


////////////////////////////////////////////////////////////////////////////////
// DESTROY SHOT
//
// The shot has collided so besides destroying the shot entity:
//    In case of hshot updates the scoreboard and clears its flag  
//    In case of vshot updates the scoreboard and number of enemies killed
//    (vshot flag will be cleared after the enemy_death_animation )
// 
// Input: Shot entity that has collided with enemy(vshot) or star(hshot)
//
//
void man_game_destroy_player_shot(Entity_t* e_shot) {
   u8 is_hshot = e_shot->type & e_type_player_hshot;

   if (is_hshot) { 
      if (e_shot->x < m_player_x)  m_player_lhshot = 0;
      if (e_shot->x > m_player_x)  m_player_rhshot = 0;
      
      man_scoreboard_set(0,0,0,0,2,0);                       // Update scoreboard +20   
   }
   else {
      man_scoreboard_set(0,0,0,0,3,1);                       // Update scoreboard +31  
   }

   man_entity_set4destruction(e_shot);
}


////////////////////////////////////////////////////////////////////////////////
// PLAYER VERTICAL SHOT
//
// Creates a shot from the player if player can shoot.
// Player only can shoot if weapon temperature bar is not full and there is no
// ongoing shot (only one shot at a time)
// If weapon temperature bar gets to max level player loses a live
//
// Input:
//    - e: Player entity
//
// TODO: RESET BAR if m_bar_lvl == MAX_BAR_LEVEL
void man_game_player_vshot(Entity_t* e_player) {
 
   if (m_player_vshot) return;

   if (m_bar_lvl == MAX_BAR_LEVEL) {                                            // Bar is full, player lose one life
      man_game_player_hit(e_player);
      return;
   }           
   
   // Vertical shot
   {
      Entity_t* e_shot = man_entity_create_from_template( &playervshot_tmpl );
      e_shot->y = e_player->y;
     
      if (StatusMatches(sys_input_status, 0x0C)) {                                // Aiming up (vertical with left hand)
         e_shot->x = e_player->x+1;
         e_shot->sprite=spr_player_vshot_rp;                                      // Right pixel
      }                                                         
      if (StatusMatches(sys_input_status, 0x0D)) {                                // Aiming up and left  (vertical with right hand)
         e_shot->x = e_player->x+6;                                              
         e_shot->sprite=spr_player_vshot_rp;                                      // Right pixel
      }
      if (StatusMatches(sys_input_status, 0x0E)) {                                // Aiming up and right (vertical with left hand)
         e_shot->x = e_player->x+1;               
         e_shot->sprite=spr_player_vshot_lp;                                      // Left pixel
      }                               

      man_game_shadow_eraser(e_shot, spr_vshot_0);                               // Shadow shot to erase trail
   }

   m_bar_lvl++;                                                                  // Bar temp goes up every vshot
   m_player_vshot = 1;                                                           // Activate player_vshot flag 

}


////////////////////////////////////////////////////////////////////////////////
// PLAYER HORIZONTAL SHOT
//
// Creates a shot from the player if player can shoot.
// Player only can shoot if weapon temperature bar is not full and there is no
// ongoing shot (only one shot at a time)
// If weapon temperature bar gets to max level player loses a live
//
// Boundaries verified by physics
//
// Input:
//    - e: Player entity
//
// TODO: RESET BAR if m_bar_lvl == MAX_BAR_LEVEL
void man_game_player_hshot(Entity_t* e_player, u8 direction) {
 
   if ((m_player_lhshot) && (direction == SHOT_LEFT))  return;
   if ((m_player_rhshot) && (direction == SHOT_RIGHT)) return;

   if (m_bar_lvl == MAX_BAR_LEVEL) {                                             // Bar is full, player lose one life
      man_game_player_hit(e_player);
      return;
   }           
   
   // Horizontal shot
   {
      Entity_t* e_shot = man_entity_create_from_template( &playerhshot_tmpl );
      
      if (m_player_avatar == HUMAN)
         e_shot->y = (StatusMatches(sys_input_status, 0x0B)) ? e_player->y + 7 : e_player->y + 8;
      else
         e_shot->y = (StatusMatches(sys_input_status, 0x0B)) ? e_player->y + 9 : e_player->y + 10;
   
      
      
      if (direction == SHOT_LEFT) {
         m_player_lhshot = 1;                                                // Activate player left horizontal shot flag 
         e_shot->x = e_player->x + 1;
         e_shot->vx = -2;
         man_game_shadow_eraser(e_shot, spr_hshot_0);                        // Shadow shot to erase trail
      }
      else {
         m_player_rhshot = 1;                                                // Activate player right horizontal shot flag 
         e_shot->x = e_player->x + 5;
         e_shot->vx =  2;         
         man_game_shadow_eraser(e_shot, spr_hshot_0);                        // Shadow shot to erase trail
      }
   }
          
   if ((((++m_num_hshot) & 3) == 0) && (!m_player_vshot)) m_bar_lvl++;       // Bar temp goes up every 4 hshot
                           
}


////////////////////////////////////////////////////////////////////////////////
// ENEMY SHOT
//
// Creates an attack (shot/bot) from the enemy if enemy can shoot or deploy a bot
// (x coordinate centered in the enemy)
//
// Input:
//    - e: Lane 0 enemy entity
//
void  man_game_enemy_attack(Entity_t* e_enemy) {
   u8 enemy_x_center = (e_enemy->x) + ((e_enemy->w) / 2);
   
   // Drop new bot only if no active bots or shots already
   if ( m_enemy_bot || m_enemy_shot ) return;                                      // Check if there is a bot or shot already  

   e_enemy->y = LANE0_Y;                                                           // In case its bouncing

   enemy_x_center += e_enemy->vx;                                                  // Since enemy moves shot x-coord adjust to current direction

   man_level_create_enemy_attack(enemy_x_center);                                  // Shot centered by the enemy x-coord

}




////////////////////////////////////////////////////////////////////////////////
// SHADOW ERASER
//
// Creates a shadow shot that follows the original erasing the trail
//
// Input:
//    - Original shot and y coordinate
//
void man_game_shadow_eraser(Entity_t* original, u8 const * spr_shadow) {
   Entity_t* clone = man_entity_clone(original);
   
   if (clone->type == e_type_enemy_bot) {
      clone->h = 4;                                      // same as bot->vy (acid is also 4)
      clone->ai_counter = 2;
}
   else if (clone->type == e_type_enemy_shot) {            // shot1 and shot2 vy is 8
      clone->h = 8;
      clone->ai_counter = 2;
   }  
   else {
      clone->ai_counter = 1; 
   }

   clone->type = e_type_empty;                        
   clone->sprite = spr_shadow;                  
   clone->collides_against = 0;                       // Collides against nothing, it will be destroyed by timer
   clone->cmps &= ~e_cmp_collider;                    // Nothing can collide with it

   clone->cmps &= ~e_cmp_movable;                     // Nothing can collide with it
   clone->cmps &= ~e_cmp_render;                      // Nothing can collide with it  

   clone->ai_behaviour = sys_ai_behaviour_render_delay;

}


////////////////////////////////////////////////////////////////////////////////
// PLAYER RESET 
//
// Reset some fields of the player as before the player death (x coordinate does not change)
//
// Input:
//    - Pointer to player entity
//
void man_game_player_reset(Entity_t* e_player) {

   man_entity_restore_template(e_player, m_player_tmpl);                             // restore template values
   e_player->x = m_player_x;                                                         // same x as before the reset

   m_player_vshot  = 0;                                                              // clear vshot flag
   m_player_lhshot = 0;                                                              // clear lhshot flag
   m_player_rhshot = 0;                                                              // clear rhshot flag

}


////////////////////////////////////////////////////////////////////////////////
// PLAYER/ENEMY DEATH ANIMATION
//
// After enemy gets hit there is some animation.  Entity still has e_cmp_collider so ghost
// shot can collide with it, but it doesn't update the points or kills again.
// No input until animation finished
//
// Input:
//    - Pointer to entity
//
/*
void man_game_death_animation(Entity_t* e, u8 duration) {

    // timer for death animation
     e->cmps = (e_cmp_movable  | e_cmp_animated | e_cmp_ai | e_cmp_render | e_cmp_collider); 
     e->ai_counter = duration;                            // time to enemy self destruction
     e->anim_counter = man_anim_enemy_death_time;

     e->ai_behaviour = sys_ai_behaviour_autodestroy;      // set self-destruction timer
 }
 */




void man_game_cleanup_attacks() {

   man_entity_forall_matching( man_entity_set4destruction, OnGoingAttack );     

}

