//-----------------------------LICENSE NOTICE------------------------------------
//
////////////////////////////////////////////////////////////////////////////////
///////////////////////////  SHOOT AND RUN /////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
//
//  Code & Graphics Copyright (C) 2020 Rafa Gutierrez (rafagutierrez@qq.com)
//
//  This file is part of Shoot and run: an Amstrad CPC Game made with CPCTelera
//  for CPCRetroDev 2020
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.


#include <man/level.h>
#include <man/entity.h>
#include <man/game.h>
#include <man/entity_templates.h>
#include <sys/ai.h>
#include <sprites/enemies_human.h>
#include <sprites/enemies_alien.h>
#include <sprites/minions_human.h>
#include <sprites/minions_alien.h>
#include <sprites/vshot.h>
#include <sprites/eggs.h>
#include <sprites/star_bot.h>
#include <sprites/rocket.h>
#include <sprites/bird_product.h>
#include <sprites/mothership.h>
#include <sprites/cloud.h>

#include <sys/render.h>

//==============================================================================
// PUBLIC DATA MEMBERS
//==============================================================================
// do not initialize global variables (SDCC issue RAM/ROM)
//level_t     level;
u8          m_level_enemies_killed;       // num of enemies killed  in current level
u8          m_level_created_enemies;      // num of enemies created in current level
u8          enemy_type;                   // Enemy type for current level

//==============================================================================
// PRIVATE DATA MEMBERS
//==============================================================================

Entity_t*            mothership;

const Entity_t*      m_lvl_enemy_tmpl;
const Entity_t*      m_lvl_minion_tmp;
const Entity_t*      m_lvl_attack_tmpl;

u8 const *           spr_mothership[2];
u8 const *           spr_enemy[4];
AnimFrame_t const *  anim_enemy[4];
AnimFrame_t const *  anim_enemy_death[4];

//==============================================================================
// PRIVATE FUNCTIONS
//==============================================================================

////////////////////////////////////////////////////////////////////////////////
// SETUP LEVEL TEMPLATES
//    Sets up the templates to use for the creation of enemies and attacks in the 
// the current level enemy type
//
void setup_level_templates() {

    // Set the enemy and attack templates for current level

    if (m_player_avatar == HUMAN) {                 // player is human
        if (enemy_type < 2) {
            m_lvl_enemy_tmpl   = (enemy_type == 0) ? &human_enemy1_tmpl     : &human_enemy2_tmpl;
            m_lvl_attack_tmpl  = (enemy_type == 0) ? &enemyshot1_tmpl       : &enemyshot2_tmpl;
            m_lvl_minion_tmp   = (enemy_type == 0) ? &human_minion1_tmpl    : &human_minion2_tmpl;
        }
        else {
            m_lvl_enemy_tmpl   = (enemy_type == 2) ? &human_enemy3_tmpl     : &human_enemy4_tmpl;
            m_lvl_attack_tmpl  = (enemy_type == 2) ? &enemyrocket_tmpl      : &enemybot_tmpl;           // dron
            m_lvl_minion_tmp   = (enemy_type == 2) ? &human_minion3_tmpl    : &human_minion4_tmpl;
        }
    }
    else {                                          // player is alien
        if (enemy_type < 2) {
            m_lvl_enemy_tmpl   = (enemy_type == 0) ? &alien_enemy1_tmpl     : &alien_enemy2_tmpl;
            m_lvl_attack_tmpl  = (enemy_type == 0) ? &eggshot1_tmpl         : &eggshot2_tmpl;           //egg
            m_lvl_minion_tmp   = (enemy_type == 0) ? &alien_minion1_tmpl    : &alien_minion2_tmpl;
        }
        else {
            m_lvl_enemy_tmpl   = (enemy_type == 2) ? &alien_enemy3_tmpl     : &alien_enemy4_tmpl;
            m_lvl_attack_tmpl  = (enemy_type == 2) ? &bird_product_tmpl     : &chickenbot_tmpl;           
            m_lvl_minion_tmp   = (enemy_type == 2) ? &alien_minion3_tmpl    : &alien_minion4_tmpl;
        }

    }    
}


//==============================================================================
// PUBLIC FUNCTIONS
//==============================================================================


u8 m_level_last_kill() {
      return (m_level_enemies_killed == MAX_ENEMIES_LVL);
}


void man_level_setup(u8 level) {

   // Clear all ongoing vshots, hshots or drones
  // man_game_cleanup_attacks();

   enemy_type              = level & 3;         // Enemy type for this level (only 4 types of enemies)
   m_level_created_enemies = 0;                 // Enemies created by the motherboard in this lvl 
   setup_level_templates();

   // Flags                                              
   m_enemy_bounces = (level > 3) ? 1 : 0;       // Enemy bouncing activated after level 4
   m_enemy_hides   = (level > 7) ? 1 : 0;       // Enemy hidden   activated after level 8


   // Mothership   
   if (level == FIRST_LVL) {
      if (m_player_avatar == HUMAN)
        mothership = man_entity_create_from_template(&mothership_tmpl);
      else
        mothership = man_entity_create_from_template(&cloud_tmpl);
   }
   else {
        
      mothership->cmps &= ~e_cmp_ai;                                                // Stop mothership from creating enemies
      wait_game_cycles(1);                                                          // Cleanup
      //wait(80);                                                                     // Pause before level ready
      mothership->cmps |= e_cmp_ai;                                                 // Mothership generate enemies again

   }
    m_level_enemies_killed  = 0;                                                    // Enemies killed by player in this lvl
      
}



////////////////////////////////////////////////////////////////////////////////
// CREATE ENEMY
//    Creates a new enemy chosing the right enemy template for the current level 
//    (only 4 types of enemies)
//           
Entity_t* man_level_create_enemy() {
 
    Entity_t* new_enemy = man_entity_create_from_template(m_lvl_enemy_tmpl);
     
    return new_enemy;
    
}


////////////////////////////////////////////////////////////////////////////////
// CREATE MINION
//    Creates a new enemy chosing the right minion template for the current level 
//    (only 4 types of enemies)
//           
Entity_t* man_level_create_minion() {
 
    Entity_t* new_minion = man_entity_create_from_template(m_lvl_minion_tmp);
     
    return new_minion;
    
}


////////////////////////////////////////////////////////////////////////////////
// LEVEL ATTACK
//    Returns the type of attack behaviour for the current level
//    (only 4 types of enemies)
//  
BehaviourFunc_t man_level_lane0_ai() {
   BehaviourFunc_t attack;

   attack = (enemy_type == 3) ? sys_ai_behaviour_deploy_bot : sys_ai_behaviour_shooter;
   
   return attack;
}



////////////////////////////////////////////////////////////////////////////////
// CREATE ENEMY SHOT
//    Creates a new enemy shot chosing the right apperance for the current level 
//    (only 4 types of enemies)
//
//  Input: x coordinate to use to center the shoot
//  
Entity_t* man_level_create_enemy_attack(u8 x) {
    Entity_t* new_shot;

    new_shot = man_entity_create_from_template( m_lvl_attack_tmpl );

    if (enemy_type < 3) m_enemy_shot = 1;              // Activate enemy_shot flag 
    else                m_enemy_bot  = 1;              // Activate enemy_bot flag 
    //(this flag only if we want to have shots and bots at the same time....)

    switch (enemy_type) {
        case 0:     new_shot->x = x;                                                    // Shot already adjust to enemy center
                    man_game_shadow_eraser(new_shot, spr_eggs_0);                      // Shadow shot to erase trail  
                    break;
        case 1:     new_shot->x = x;                                                    // Shot already adjust to enemy center
                    man_game_shadow_eraser(new_shot, spr_eggs_0);                      // Shadow shot to erase trail  
                    break;
        case 2:     x -= (new_shot->w)/2;                                               // Shot center adjust to enemy center
                    new_shot->x = x;  
                    man_game_shadow_eraser(new_shot, spr_rocket_0);                      // Shadow shot to erase trail  
                    break;
        case 3:     x -= (new_shot->w)/2;                                               // Shot center adjust to enemy center
                    new_shot->x = x;  
                    man_game_shadow_eraser(new_shot, spr_star_bot_0);                   // Shadow shot to erase trail   
    }                 
 
    return new_shot;

}


////////////////////////////////////////////////////////////////////////////////
// SET ENEMY DEATH
//    Setup the proper death animation for the type of enemy
//
void man_level_set_enemy_death(Entity_t* e) {
    AnimFrame_t const *  anim_human_enemy_death[4] = {
        man_anim_human_enemy1_death, man_anim_human_enemy2_death, man_anim_human_enemy3_death, man_anim_human_enemy4_death
    };
    AnimFrame_t const *  anim_human_minion_death[4] = {
        man_anim_human_minion1_death, man_anim_human_minion2_death, man_anim_human_minion3_death, man_anim_human_minion4_death
    };
    AnimFrame_t const *  anim_alien_enemy_death[4] = {
        man_anim_alien_enemy1_death, man_anim_alien_enemy2_death, man_anim_alien_enemy3_death, man_anim_alien_enemy4_death
    };
    AnimFrame_t const *  anim_alien_minion_death[4] = {
        man_anim_alien_minion1_death, man_anim_alien_minion2_death, man_anim_alien_minion3_death, man_anim_alien_minion4_death
    };


    if (m_player_avatar == HUMAN) {                 // player is human
        if (m_level < SPLIT_LEVEL) 
            e->anim   = anim_human_enemy_death[enemy_type]; 
        else 
            e->anim   = anim_human_minion_death[enemy_type]; 
    }
    else {                                          // player is alien
        if (m_level < SPLIT_LEVEL) 
            e->anim   = anim_alien_enemy_death[enemy_type]; 
        else 
            e->anim   = anim_alien_minion_death[enemy_type]; 
    }

}

// This option does not require a template for each type of enemy
//
/*
void level_enemy_setup(Entity_t* e ) {
    u8 enemy_type = m_level & 3;                                

    u8 const * spr_enemy[4] = {
        spr_enemy0_0, spr_enemy1_0, spr_enemy2_0, spr_enemy3_0
    };
    
    AnimFrame_t const *  anim_enemy[4] = {
        man_anim_enemy0, man_anim_enemy1, man_anim_enemy2, man_anim_enemy3
    };       

    e->sprite = spr_enemy[enemy_type]; 
    e->anim   = anim_enemy[enemy_type]; 
}
*/


// WHAT IT NEEDS TO BE DONE EVERY TIME WE START A LEVEL 
// IS NOT THE SAME AS
// WHAT IT NEEDS TO BE DONE EVERY TIME WE START A GAME
/*
void level_init() {
   level.num = level;                                           // levels: 0..11

   level.enemies_killed = 0;
   
   level.enemies_created = 0;                                   // Enemies created by the motherboard in this lvl

   level.id_palette = ((level & 1)==1) ? 0 : 1;
   

    // Enemies
   {
      u8 enemy_type = level & 3;                                // There are only 4 types of enemies

      // set the right sprite and death animation so when we create a new enemy it uses the right sprites
      if      (enemy_type & 0) level_set_enemies(0);
      else if (enemy_type & 1) level_set_enemies(1);
      else if (enemy_type & 2) level_set_enemies(2);
      else if (enemy_type & 3) level_set_enemies(3);
   }

   // Enemy bouncing 
   level.bouncing_on = (level > 3 ) ? 1 : 0;
   
   // Enemy hidding
   level.hidding_on = (level > 7 ) ? 1 : 0;                     // not implemented yet
    

   ////////////////////////////////////////////////////////////////////////////////////////////////
   ////////////////////////////////// THINK ABOUT IT //////////////////////////////////////////////
   ////////////////////////////////////////////////////////////////////////////////////////////////

   // In game.c initialize
   //    
   // Entities                          Need to delete all previous enemies....
   // Scoreboard                        It does not change from one level to the next one
   // Weapon temperature bar            It does not change from one level to the next one
   // Player livesboard                 It does not change from one level to the next one

   // No shots on start
   m_player_vshot = 0;
   m_player_hshot = 0;
   m_enemy_shot   = 0;
   m_enemy_bot    = 0;

   // No enemies on any lane on start
   cpct_memset(m_lane_status, 0, sizeof(m_lane_status));


   // Player
   m_player_x = PLAYER_X;
   man_entity_create_from_template(&player_tmpl);

   // Systems that need initialization
   sys_ai_init();
   sys_render_init(); 

}
*/