//-----------------------------LICENSE NOTICE------------------------------------
//
////////////////////////////////////////////////////////////////////////////////
///////////////////////////  SHOOT AND RUN /////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
//
//  Code & Graphics Copyright (C) 2020 Rafa Gutierrez (rafagutierrez@qq.com)
//
//  This file is part of Shoot and run: an Amstrad CPC Game made with CPCTelera
//  for CPCRetroDev 2020
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.


#include "ai.h"
#include <sys/input.h>
#include <sprites/mothership.h>                        // review
#include <sprites/cloud.h>                        // review
#include <sprites/light.h>
#include <sprites/bar_levels.h>
#include <sys/render.h>
#include <man/entity.h>
#include <man/game.h>
#include <man/level.h>
#include <man/livesboard.h>
#include <man/entity_templates.h>



/// REVIEW SOME FUNCTIONS WITH TIMER TO SEE IF THEY CAN BE ADDED TO THE GENERAL 
//  FUNCTION sys_ai_behaviour_autodestroy(Entity_t* e)

u8 b_step;                                              // mothership bouncing step
i8 const bouncing[]={-1, -2, -1, 0, 1, 2, 1, 0};        // mothership bouncing table (gentle)

u8 f_step;                                              // Sky FX flash step
i8 const flash[]={
    HW_BRIGHT_WHITE,    HW_BRIGHT_WHITE, HW_BRIGHT_WHITE,
    HW_PASTEL_YELLOW,   HW_ORANGE, 
    HW_BRIGHT_RED,      HW_BRIGHT_RED,
    HW_RED,             HW_YELLOW,
    HW_CYAN,            HW_BRIGHT_BLUE,
    HW_SKY_BLUE,        HW_PASTEL_BLUE,
    HW_SKY_BLUE,        HW_CYAN   
};

u8 hidden_counter;                                      // Timer for hidden/unhidden enemies



//==============================================================================
//==============================================================================
// PRIVATE FUNCTIONS
//==============================================================================
//==============================================================================

////////////////////////////////////////////////////////////////////////////////
// UPDATE ONE ENTITY
//      Updates physics of a given entity
// Inputs:
//      e: Pointer to the entity to be updated
//
// Since we use the signature in the call from forall_matching in ai_update, 
// we know that all the entities passed as parameters have  e_cmp_ai = 1
void sys_ai_update_one_entity(Entity_t* e) {
    e->ai_behaviour(e);
}


////////////////////////////////////////////////////////////////////////////////
// LEFT-RIGHT BEHAVIOUR
//      Changes velocity of the entity when touching lef or right boundaries
// Inputs:
//      e: Pointer to the entity
//
void sys_ai_behaviour_left_right(Entity_t* e, u8 left_limit, u8 right_limit) {
    u8 new_x;

    new_x = e->x + e->vx;
    // check boundaries
    if (new_x        <   left_limit)    e->vx =  1;
    if (new_x + e->w >   right_limit)   e->vx = -1;

}


////////////////////////////////////////////////////////////////////////////////
// BOUNCING BEHAVIOUR
//      Bounces the entity following the bouncing table
//
// Inputs:
//      e: Pointer to the entity
//      y: Y coordinate used as a base to calculate the bouncing from.
//
void sys_ai_behaviour_bouncing(Entity_t* e, u8 y) {

    if ((e->x & 1) == 0) {                            
       if ((e->type & e_type_empty)) b_step++;               // Only mothership updates the step
       b_step &= 7;                                          // step = [0..7]
       e->y = y + bouncing[b_step];                          // Bouncing table has 8 elements
    }

}



////////////////////////////////////////////////////////////////////////////////
// ENEMY BOUNCING BEHAVIOUR
//
// Inputs:
//      e: Pointer to an entity in lane 0, lane 1 or lane 2 
//
void sys_ai_behaviour_bouncing_enemy(Entity_t* e) {

   if       ( e->y > LANE1_Y + 2) sys_ai_behaviour_bouncing(e, LANE0_Y);
   else if  ( e->y > LANE2_Y + 2) sys_ai_behaviour_bouncing(e, LANE1_Y);
   else                           sys_ai_behaviour_bouncing(e, LANE2_Y);
   
}



////////////////////////////////////////////////////////////////////////////////
// CHANGES DIRECTION BEHAVIOUR
//      Changes velocity of the entity depends the attitude of the entity
// Inputs:
//      e: Pointer to the entity
//      attitude: aggressive/defensive
//
 void sys_ai_behaviour_change_direction(Entity_t* e, u8 target_x, u8 attitude) {
    
    switch (attitude) {
        case SYS_AI_AGGRESSIVE: e->vx = (target_x > e->x) ? 1 : -1;
                                break;
        case SYS_AI_DEFENSIVE:  e->vx = (target_x < e->x) ? 1 : -1;
                                break;
        // REVIEW:
        default:                if (cpct_rand() < 30) e->vx = -e->vx;                
    }
    
}


void sys_ai_behaviour_light(Entity_t* e) {
 
    if (e->y == (LIGHT_Y + LIGHT_H)) {
        e->sprite = spr_light_0;
        e->vy = (e->vy == 0) ? -2 : 0;
        m_abduction_light = 2;
    }

    if ((e->y <= (LIGHT_Y+2)) && (e->vy < 0)) {      
        man_game_entity_destroy(e);
        m_abduction_light = 0;
    }
    
}


////////////////////////////////////////////////////////////////////////////////
// FROZEN MOTHERSHIP 
//
//
void sys_ai_frozen_mothership(Entity_t* e) {                                              

    if (e->ai_counter == 0)   {                                     // counter is off
       e->ai_counter = SYS_AI_MOTHERSHIP_FROZEN;                    // start the counter
       e->cmps &= ~e_cmp_movable;                                   // stop movement
       e->cmps &= ~e_cmp_animated;                                   // stop animation
       e->sprite = (m_player_avatar == HUMAN) ? spr_mothership_2 : spr_cloud_0;
       
       
       // Create entity lights
       // Note: lights blink cause mothership sprite overlaps a bit and keeps refreshing     
       {           
            Entity_t* lights;
            if (m_player_avatar == HUMAN)
                lights = man_entity_create_from_template(&mothership_light_tmpl);
            else
                lights = man_entity_create_from_template(&cloud_light_tmpl);

            lights->x = e->x + 5;
            lights->y = e->y + e->h - 1;
            lights->ai_counter = 8;                                 // Autodestroy timer 
       }
       
     }

    if ((--e->ai_counter) == 0) {                                 // time's up! start movement
        e->cmps |= (e_cmp_movable | e_cmp_animated);         
        sys_ai_behaviour_left_right(e, 1, 79);
    }

}


////////////////////////////////////////////////////////////////////////////////
// HIDE/UNHIDE ALL ENEMIES AT THE SAME TIME
//
//
void sys_ai_switch_render(Entity_t* e) {

if (hidden_counter < 15) {                      // Time is up
       Entity_t* c = man_entity_clone(e);       // Delete enemy last trail with a ghost
       c->cmps = e_cmp_render;
       man_entity_set4destruction(c);
    
       e->cmps &= ~e_cmp_render;                // Hidde entity  
    }
else 
    e->cmps |= e_cmp_render;                    // Show entity       

}


//==============================================================================
//==============================================================================
// PUBLIC FUNCTIONS
//==============================================================================
//==============================================================================

void sys_ai_init() {
    b_step = 0;
    f_step = 0;
    hidden_counter= 0;
}

////////////////////////////////////////////////////////////////////////////////
// UPDATE
//      Updates the AI system
//
// Control inversion:
// The system call the entity manager that loops through all the entities
// calling the ai update for one entity.
// (we provide a pointer to the function to update the ai of one entity)
//
// The system just moves the ai entities that can be moved (signature movable bit = 1)
// Matching the signature we avoid calling the function for entities without ai or can't be moved.
void sys_ai_update() {

    man_entity_forall_matching( sys_ai_update_one_entity,  e_cmp_ai );

}



//==============================================================================
//==============================================================================
// PUBLIC BEHAVIOURS
//==============================================================================
//==============================================================================


void sys_ai_behaviour_enemy_splits(Entity_t* e) {
    Entity_t* minion_left  = 0; 
    Entity_t* minion_right = 0; 
    u8 lane = man_game_check_lane(e->y);

    // Create minion left
    // THIS TEMPLATE NEEDS TO CHECK WITH LEVEL MANAGER TO KNOW WHAT TYPE OF MINION TEMPLATE IS NEEDED
    minion_left = man_level_create_minion();
    minion_left->x = e->x;
    minion_left->y = e->y;
    minion_left->vx = -1;
    if (lane == 0) minion_left->ai_behaviour = man_level_lane0_ai();

    // Create minion right
    minion_right = man_entity_clone(minion_left);
    minion_right->x += 6;
    minion_right->vx = 1; 
       
    m_lane_status[lane] = 2;                    // Lane now has two minions
    
    e->cmps = e_cmp_render;                     // Original entity dies
    man_entity_set4destruction(e);

    m_player_vshot = 0;                         // Player can shot again

}


////////////////////////////////////////////////////////////////////////////////
// ENEMY BEHAVIOUR WHEN GETS FINAL HIT
//
// Enemy behaviour when it gets hit: sets death animation + self-destruction timer
//
// Inputs:
//      e: Pointer to the entity of the enemy
//
void sys_ai_behaviour_enemy_dies(Entity_t* e) {
   // mothership frozen
   // mothership change sprite
   // player frozen
    
   man_level_set_enemy_death(e);                            // sets the proper death animation 

   // timer for death animation
   e->cmps = (e_cmp_movable  | e_cmp_animated | e_cmp_ai | e_cmp_render | e_cmp_collider); 
   e->ai_counter = SYS_AI_ENEMY_DEATH;                      // time to enemy self destruction
   e->anim_counter = man_anim_enemy_death_time;
    
   e->ai_behaviour = sys_ai_behaviour_autodestroy_by_t;          // set self-destruction timer
  
}


////////////////////////////////////////////////////////////////////////////////
// PLAYER BEHAVIOUR WHEN GETS HIT
//
// Player behaviour when it gets hit: sets death animation + self-destruction timer
//
// Inputs:
//      e: Pointer to the entity of the player
//
void sys_ai_behaviour_player_dies(Entity_t* e_player) {

    e_player->anim = (m_player_avatar == HUMAN) ? man_anim_player_death : man_anim_alien_death;

     // timer for death animation
    e_player->vx = 0;
    e_player->cmps = (e_cmp_movable  | e_cmp_animated | e_cmp_ai | e_cmp_render | e_cmp_collider); 
    e_player->ai_counter = SYS_AI_PLAYER_DEATH;                     // time to enemy self destruction
    e_player->anim_counter = man_anim_enemy_death_time;

    e_player->ai_behaviour = sys_ai_behaviour_autodestroy_by_t;     // set self-destruction timer

}


////////////////////////////////////////////////////////////////////////////////
// ENEMY SHOOTER BEHAVIOUR
//
// It tracks the player and shot when in range
//
// Inputs:
//      e: Pointer to the entity in lane 0
//
void sys_ai_behaviour_shooter(Entity_t* e) {  
   u8 target_x = e->x + 2;       
   u8 range = (m_player_x > target_x) ? m_player_x - target_x : target_x - m_player_x;

   if ((e->x == 10) || (e->x == 65)) sys_ai_behaviour_change_direction(e, m_player_x, SYS_AI_AGGRESSIVE);

    if (m_difficulty == 0) {                                              // By position
        if ((e->x < 65) && ((e->x &15) == 0)) man_game_enemy_attack(e);
    }
    else {                                                                // By range
        if ( range < 4) man_game_enemy_attack(e);
    }

   sys_ai_behaviour_left_right(e, 1, 79);

   if (m_enemy_bounces) sys_ai_behaviour_bouncing_enemy(e);

   if (m_enemy_hides) sys_ai_switch_render(e);

   m_lane0_enemy_x = e->x;          // Update lane0 enemy x coord
}



////////////////////////////////////////////////////////////////////////////////
// ENEMY DEPLOY BOT BEHAVIOUR
//
// It tracks the player and shot when in range
//
// Inputs:
//      e: Pointer to the entity in lane 0
//
void sys_ai_behaviour_deploy_bot(Entity_t* e) {  

   sys_ai_behaviour_left_right(e, 1, 79);  


    if (m_difficulty < 2) {
        if (((e->x > 0) && (e->x < 15)) || ((e->x > 55) && (e->x < 70))) man_game_enemy_attack(e);
    }
    else {
        if ((e->x < 65) && ((e->x &15) == 0)) man_game_enemy_attack(e);
    }

   if (m_enemy_bounces) sys_ai_behaviour_bouncing_enemy(e);

   if (m_enemy_hides) sys_ai_switch_render(e);

   m_lane0_enemy_x = e->x;          // Update lane0 enemy x coord

}


////////////////////////////////////////////////////////////////////////////////
// ENEMY BEHAVIOUR
// Inputs:
//      e: Pointer to the entity 
//
void sys_ai_behaviour_enemy(Entity_t* e) {

    if ((e->x & 0x0F) == 0) sys_ai_behaviour_change_direction(e, m_player_x, SYS_AI_REGULAR);   
    sys_ai_behaviour_left_right(e, 1, 79);

    if (( e->x & 0x07) == 0) man_game_enemy_lane_down(e);  // Multiple of 8: remainder (e->x & 0b00000111) == 0

    if (m_enemy_bounces) sys_ai_behaviour_bouncing_enemy(e);   

    if (m_enemy_hides) sys_ai_switch_render(e);
    
}


void sys_ai_behaviour_player_zero(Entity_t* e_player) {
    u8 enemy_type = m_level & 3;
    u8 new_x      = e_player->x + e_player->vx;

 // REVIEW!!!!
 /*  
    e_player->vx = 0;

    if (m_star != 0) {                                          // 0: no star 2: star on right of player 1: star on left of player
        sys_input_status |= 0x08;
        sys_input_status |= (m_star == 1) ? 0x01 : 0x02;
        man_game_player_hshot(e_player, m_star);
        sys_physics_update_velocity(e_player); 
        sys_physics_update_shots_status(e_player);
        m_star = 0;
    }
*/
/*
    if (m_lane0_enemy_x != 0) {                                             // There is an enemy in lane0
        u8 target_x = e_player->x;
        u8 left_side_shooter =  m_lane0_enemy_x < target_x;
        u8 range =  left_side_shooter ? target_x - m_lane0_enemy_x : m_lane0_enemy_x - target_x;

        // Movement
        if (enemy_type != 3) {             // no bots
            if ((!m_enemy_shot) && (new_x > 10) && (new_x < 63)) 
                sys_ai_behaviour_change_direction(e_player, m_lane0_enemy_x, SYS_AI_AGGRESSIVE);
            
            if ((m_enemy_shot) && (new_x > 10) && (new_x < 63) && (range < 12)) 
                sys_ai_behaviour_change_direction(e_player, m_lane0_enemy_x, SYS_AI_DEFENSIVE);
            sys_input_status |= (e_player->vx == -1) ? 0x01 : 0x02;
        }
       else {                             // bots
            if (m_enemy_bot) {                                              // A bot is on its way...
                if (left_side_shooter)                                      // Enemy drop a bot on the left side of the player
                    e_player->vx = (new_x + e_player->w > 63) ? -1 : 1;     // Player on right limit->stops, otherwise move to the right
                else                                                        // Enemy drop a bot on the right side of the player
                    e_player->vx = (new_x < 10) ? 1 : -1;                   // Player on left limit->stops, otherwise move to the left
                sys_input_status |= (e_player->vx == -1) ? 0x01 : 0x02;
            }

        }

        // Shoot by range
        if ( range < 8) 
            sys_input_status |= 0x08;
            sys_input_status |= 0x04;
            man_game_player_vshot(e_player);

        if (m_star != 0) {                                          // 0: no star 2: star on right of player 1: star on left of player
            sys_input_status |= 0x08;
            sys_input_status |= (m_star == 1) ? 0x01 : 0x02;
            man_game_player_hshot(e_player, m_star);
            m_star = 0;
        }
            
    }
*/

}



////////////////////////////////////////////////////////////////////////////////
// MOTHERSHIP BEHAVIOUR
//     Baisc behaviour of the mothership, moving left-right and occasionally 
// new enemies 
// Inputs:
//      e: Pointer to the entity 
//
void sys_ai_behaviour_mothership(Entity_t* e) {

    hidden_counter++;
    hidden_counter &= 31;

    if (m_level_created_enemies < MAX_ENEMIES_LVL) {                   // still need to create more enemies

         if (m_lane_status[2] == 0) {                                  // lane2 free -> create a new enemy
            sys_ai_frozen_mothership(e);                               // stop movement
            man_game_create_enemy(e);                                  // new enemy
            m_level_created_enemies++;
            m_lane_status[2] = 1;                                      // Lane 2 now has 1 enemy
            return;
         }
    }

    if (e->ai_counter == 0) {                                           // normal behaviour
        sys_ai_behaviour_bouncing(e, MOTHERSHIP_Y);    
        sys_ai_behaviour_left_right(e, 1, 79);
    }
    else 
        sys_ai_frozen_mothership(e);                                    // check counter
     
}


////////////////////////////////////////////////////////////////////////////////
// MOTHERSHIP ABDUCES BEHAVIOUR
//     Mothership moves left and right searching for the player, once found it
// turns on the abduction light (then player cannot move) and capture the player
// Inputs:
//      e: Pointer to the entity 
//
void sys_ai_behaviour_mothership_abduces(Entity_t* e) {
    u8 mothership_x      = e->x ;
    u8 target_x          = (m_player_x > 10) ? m_player_x - 6 : 0;
    
    if (e->x == target_x) {                                                         // abduction target found    

        if (e->ai_counter == 0)   {                                                 // counter is off
           e->ai_counter = LIGHT_H + 10;                                            // start the counter
           e->cmps &= ~e_cmp_movable;                                               // stop movement
           e->y = 16;

           m_abduction_light = 1;
           // Create light
           {
               Entity_t* light = man_entity_create_from_template(&light_tmpl);
               light->x = m_player_x;
               light->y = LIGHT_Y;
           }
        }

        if ((--e->ai_counter) == 0) {                                               // time's up! start movement
           e->cmps |= e_cmp_movable;         
           sys_ai_behaviour_left_right(e, 1, 79);
        }

    }
    else {                                                                        // Normal behaviour   
        if (e->ai_counter == 0) {
            e->cmps |= e_cmp_movable;        
            sys_ai_behaviour_bouncing(e, MOTHERSHIP_Y);    
            sys_ai_behaviour_left_right(e, 1, 79);
        }
        else
            --e->ai_counter;
    }
    
    hidden_counter++;
    hidden_counter &= 31;
}



////////////////////////////////////////////////////////////////////////////////
// FLASH BEHAVIOUR
// Inputs:
//      e: Pointer to the entity 
//
void sys_ai_behaviour_flash(Entity_t* e) {

    if (e->ai_counter == 0) {
        f_step = 0;
        sys_render_setBackground(HW_BLACK);
        man_entity_destroy(e);                                        // Can't call set4destruction since it doesn't have render_cmp
    }
    else {
        e->ai_counter--;
        
        if ((e->ai_counter & 1) == 0) {
           ++f_step;
           f_step &= 7;                                              // Rotate through the 15 colors
           sys_render_setBackground(flash[f_step]);
        }
        else 
            sys_render_setBackground(HW_BLACK);
    }

}


////////////////////////////////////////////////////////////////////////////////
// AUTODESTROY BEHAVIOUR
//      Destroys entity when timer reaches zero, when times up perform the 
// required behaviour for the type of entity 
//
//
// Inputs:
//      e: Pointer to the entity (e->ai_counter >=0)
//
void sys_ai_behaviour_autodestroy_by_t(Entity_t* e) {

    if (--(e->ai_counter) == 0 ) man_game_entity_destroy(e);
    else {
        // When an enemy is in death animation mode it keeps moving
        if (e->type == e_type_enemy) {
            sys_ai_behaviour_left_right(e, 1, 79);  

            if (m_enemy_bounces) sys_ai_behaviour_bouncing_enemy(e);        // need it?

            if (m_enemy_hides)   sys_ai_switch_render(e);     

            return;
        }
    }

}

void sys_ai_behaviour_autodestroy_by_x(Entity_t* e) {
    u8 e_left_x  = e->x;
    u8 e_right_x = e_left_x + e->w;

    // Check boundaries
    if (e_left_x  < 2)   man_game_entity_destroy(e);
    if (e_right_x > 77)  man_game_entity_destroy(e);

}


////////////////////////////////////////////////////////////////////////////////
// DELAY THE RENDER OF AN ENTITY
//
// Inputs:
//      e: Pointer to the entity (e->ai_counter >=0)
//
void sys_ai_behaviour_render_delay(Entity_t* e) {

    if ( --(e->ai_counter) == 0) {
        e->cmps |= (e_cmp_render | e_cmp_movable);               // time's up, render component activated
        
        if (e->vx != 0) {                                               // horizontal shots
            e->ai_behaviour = sys_ai_behaviour_autodestroy_by_x;   
            return;
        }
        if (e->vy < 0) {        
            e->ai_behaviour = sys_ai_behaviour_autodestroy_by_t;        // vertical shots
            e->ai_counter = 15;    
            return; 
        }                       
        if (e->vy > 0) e->ai_behaviour = sys_ai_behaviour_bot;           // bot

    }
}



////////////////////////////////////////////////////////////////////////////////
// BOT BEHAVIOUR
//      Bot behaviour...
//
// Inputs:
//      e: Pointer to the entity 
//
void sys_ai_behaviour_bot(Entity_t* e) {

    if  (e->type & e_type_empty ) {                                 // shadow
        if (e->y >= (GROUND_Y - 15) ) man_game_entity_destroy(e);   // destroy shadow (>= in case is bouncing)
    }
    else {
        if (e->y == GROUND_Y - e->h) {                              // bot on the ground
            e->y = GROUND_Y - e->h;                                 // In case is bouncing
            e->cmps |= e_cmp_animated;                              // starts animation of the star
            e->vy  = 0;                                             // no more falling
            e->vx  = (m_player_x > e->x) ? 1 : -1;                  // moves toward player
            if (m_star == 0)
                m_star = (e->vx == -1) ? 2 : 1;                     // 0: no star 2: star on right of player 1: star on left of player
        }
    } 
}

/*
void sys_ai_behaviour_lower_bar(Entity_t* e) {

    e->x = TEMP_BAR_X + (2*m_bar_lvl);

    if (m_bar_lvl > 0) {                                // cold down the bar
        m_bar_lvl--;
        e->sprite = spr_bar_level_0;                    // black    
        e->ai_behaviour = sys_ai_behaviour_lower_bar;           
    }
    else
        e->ai_behaviour = sys_ai_behaviour_bar;

}
*/

////////////////////////////////////////////////////////////////////////////////
// WEAPON TEMPERATURE BAR
//
// The temperature bar is an entity with 3 possible 8x8 pixels sprites (bar segments): 
// one green, one red, one black 
//
// In order to keep the render system simple, the ai will update the 
// sprite to render (black to decrease, green/red to increase) and the x coordinate.
//
// It also keeps ongoing timers when the m_bar_lvl > 0, when timer time's up
// decrease the bar (m_bar_lvl-- and black sprite so render erase current segment)
//
// Precondition: m_bar_lvl values 0..MAX_BAR_LEVEL
//
// Inputs:
//      e: Pointer to the entity 
//

void sys_ai_behaviour_bar(Entity_t* e) {

    if (man_livesboard_check() == 0) return;        // game over: no need to update bar

    // m_bar_lvl starts from 0
    // COLOR
    e->x = TEMP_BAR_X + (2*m_bar_lvl);

    if (m_bar_lvl > 9) 
        e->sprite = spr_bar_level_2;                // red
    else
        e->sprite = spr_bar_level_1;                // green

    // POSITION
    if (m_bar_lvl > 0) {
        
        if ( e->ai_counter == 0 )                   // counter is off                           
           e->ai_counter = SYS_AI_BAR_TIMER;        // start the counter
        else                                        // counter is on
           e->ai_counter--;                         // decrease counter  
        
        if ( e->ai_counter == 0 ) {                 // time's up
           m_bar_lvl--;                             // decrease the bar
           e->sprite = spr_bar_level_0;             // black                 
        }

    }
       
}






/////////////////////////////////////////////////////////////////////////////////////
// Note: To check if x is multiple of 8 we can check (x % 8) == 0 BUT IT IS TOO SLOW
//      Alternative:
//      To divide a binary number by 8 we just need to shift it to the right 3 times
//      The remainder of the division are the bits we take out from the right side
//
//      Ex:             19 = 0b00010011
//          3x Shift right = 0b00000010 011 = Remainder = 011 (3 shifted out bits)
//          If the remainder is zero means original number is multiple
//          So we just need to mask the last 3 bits with an arithmetic AND (&)
//          0b00000111 = 0x7 = 7
//          The result of the mask is the remainder, just need to check if it is zero