//-----------------------------LICENSE NOTICE------------------------------------
//
////////////////////////////////////////////////////////////////////////////////
///////////////////////////  SHOOT AND RUN /////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
//
//  Code & Graphics Copyright (C) 2020 Rafa Gutierrez (rafagutierrez@qq.com)
//
//  This file is part of Shoot and run: an Amstrad CPC Game made with CPCTelera
//  for CPCRetroDev 2020
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.


#include "collision.h"
#include <man/scoreboard.h>
#include <man/livesboard.h>
#include <man/game.h>
#include <man/entity_templates.h>


//==============================================================================
//==============================================================================
// PRIVATE FUNCTIONS
//==============================================================================
//==============================================================================

////////////////////////////////////////////////////////////////////////////////
// COLLIDED ENTITY
//
// Checks the type of the entity and do what the entity should do after a collision. 
// Most entities just get destroyed but some might need to do a few extra actions
// before getting destroyed
//
// Player/Enemy have an an animation before they die, sys_ai_behaviour_autodestroy 
// controls the duration of that animation, once timer is up it will call the 
// function man_game_entity_destroy
//
// Inputs:
//      e: Pointer to the collided entity
//
void sys_collition_collided_entity(Entity_t* e) {
    
    // Check all the types that have the potentital to collide
    if (e->type & e_type_player) {                       // It has death animation, not destroy yet
       man_game_player_hit(e);                           // Actions BEFORE destroy: animation, stop enemies attacks
       return;
    }

    if (e->type & e_type_enemy) {                        // It has death animation, not destroy yet
        man_game_enemy_hit(e);
        return;
    }

    if (e->type & e_type_enemy_bot) {
       man_game_bot_hit(e);                              // Let enemy deploy bot again and destroys bot
       return;
    }   

    if (e->type & IS_PLAYER_SHOT) {
       man_game_destroy_player_shot(e);                  
       return;
    }

    // Note: Enemies shots can only collide with player, in that case they get destroyed right away but the enemy_shot flag is
    // cleared after the player animation is finished (by the player_killed function)
    if (e->type & e_type_enemy_shot) {
        man_entity_set4destruction(e);
        return;
    }

    man_game_entity_destroy(e);                          // Any other types will get destroy here
    
}


////////////////////////////////////////////////////////////////////////////////
// UPDATE ONE ENTITY
//      Updates collisions of a given pair of entities
// Inputs:
//      e1: Pointer to the first entity
//      e2: Pointer to the second entity
//
// Since we use the signature e_cmps_collider in the call from forall_pairs_matching 
// in collition_update, we know that all the entities passed as parameters have 
// e_cmp_collider = 1 and e1 != e2
//
// e1->collides_against is a signature (bitset) of types against it collides with
//
// Ex: e2 type player:                              0b00000001
//     e2 type player:                              0b00000100
//     e1 collider collides_against player:         0b00000001
//     e1 collides_against player and mothership:   0b00000101  
// 
// If arithmetic AND between e1->collides_against and e2->type is not zero
// means that at least they both have a bit in common set to 1 therefore e1 collides against e2 
//
// Collision property is not commutative:
//       A bullet can collide against an enemy BUT an enemy does NOT collide against a bullet
//       An enemy RECEIVE COLLISIONS FROM a bullet
//
void sys_collition_update_entities(Entity_t* e1, Entity_t* e2) {
    u8 types = e1->type | e2->type;

    // Check if both entities can collide
    u8 e1_col_e2 = e1->collides_against & e2->type;     // non zero = e1 collides against e2
    u8 e2_col_e1 = e2->collides_against & e1->type;     // non zero = e2 collides against e1
    if ((e1_col_e2 | e2_col_e1) == 0) return;           // there is no collision

    // Entities can collide (but maybe there is no collision yet)
    // Check collision between e1 and e2 using bounding boxes
    if (VerticalOverlap(e1, e2) && (HorizontalOverlap(e1, e2))) {   

// TEMPORAL: Minions do not overlap, they just avoid each other
// Using this in minion template need to set that collides against enemies, otherwise minion collides_against nothing
//        if ((e1->type & e_type_enemy) && (e2->type & e_type_enemy)) {
//            e1->vx = -e1->vx;
//            e2->vx = -e2->vx;
//        }
//        else {
        // There is collision, check entity types and react
        sys_collition_collided_entity(e1);
        sys_collition_collided_entity(e2);

//        }
    }
}
 
////////////////////////////////////////////////////////////////////////////////
//  
//  Potential collisions
//  a)  Player, Enemy_shot
//  b)  Player, Bot
//  c)  Enemy, Player_vshot
//  d)  Bot, Player_vshot
//  e)  Bot, m_player_hshot
//
////////////////////////////////////////////////////////////////////////////////
//
//  Actions
//  a. Stop enemies from keep attacking
//     Player performs death_animation
//     After animation player can shot again (m_player_vshot = 0, m_player_hshot = 0) and enemy can shot again (m_enemy_shot = 0)
//     Player loses a life
//     If player lives > 0 reset player    
//     Otherwise           reset the game
//     Enemy_shot gets destroyed
//     
//  b. Player performs death animation
//     After animation player can shot again (m_player_vshot = 0, m_player_hshot = 0) and enemy can deploy again (m_enemy_bot = 0)
//     Player loses a life
//     If player lives > 0 reset player, otherwise reset the game
//     Bot gets destroyed
//  
//  c. Score increases by 21 points
//     Enemy performs a death animation
//     After animation player can shot again (m_player_vshot = 0)
//     Enemy gets destroyed
//  
//  d. Score increases by 10 points
//     Player can shot again (m_player_vshot = 0) and enemy can deploy a bot again (m_enemy_bot = 0)
//     Bot gets destroyed
//  
//  e. Score increases by 10 points
//     Player can shot again (m_player_hshot = 0) and enemy can deploy a bot again (m_enemy_bot = 0)
//     Bot gets destroyed


//==============================================================================
//==============================================================================
// PUBLIC FUNCTIONS
//==============================================================================
//==============================================================================

////////////////////////////////////////////////////////////////////////////////
// UPDATE
//      Updates the collitions system
//
// Control inversion:
// The system call the entity manager that loops through all the entities
// calling the collitions update for one entity.
// (we provide a pointer to the function to update the collitions of one entity)
//
// The system just checks the entities that can collide (signature collider bit = 1)
// Matching the signature we avoid calling the function for entities that can't collide.
void sys_collition_update() {
    man_entity_forall_pairs_matching( sys_collition_update_entities, e_cmp_collider );
}