//-----------------------------LICENSE NOTICE------------------------------------
//
////////////////////////////////////////////////////////////////////////////////
///////////////////////////  SHOOT AND RUN /////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
//
//  Code & Graphics Copyright (C) 2020 Rafa Gutierrez (rafagutierrez@qq.com)
//
//  This file is part of Shoot and run: an Amstrad CPC Game made with CPCTelera
//  for CPCRetroDev 2020
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.


#include "physics.h"
#include <sys/input.h>
#include <man/game.h>
#include <man/livesboard.h>


//==============================================================================
//==============================================================================
// PRIVATE FUNCTIONS
//==============================================================================
//==============================================================================

// 0000 no key pressed
// 0001 O key  pressed
// 0010 P key  pressed
// 0100 Q key  pressed
// 1000 Space key pressed
void sys_physics_update_velocity(Entity_t* e) {
    u8 new_x;

    // Left - Right Movement
    e->vx = 0;
    if      (StatusMatches(sys_input_status, 0x01)) e->vx=  -1;
    else if (StatusMatches(sys_input_status, 0x02)) e->vx =  1;
    
    if      (StatusMatches(sys_input_status, 0x03)) e->vx =  0;

    new_x = e->x + e->vx;

    // Check boundaries
    if (new_x        < LEFT_LIMIT)    e->vx = 0;
    if (new_x + e->w > RIGHT_LIMIT)   e->vx = 0;

    // Jump
    // Jump straight up:         0100
    // Jump to the right:        0101
    // Jump to the left:         0110
    // To do...

    // Update the global player x coordinate (only player has input)
    m_player_x = e->x + e->vx;

}


void sys_physics_update_shots_status(Entity_t* e) {

    // Player shots...
    if (sys_input_status & 0x08) {
        e->vx = 0;
        
        if ((sys_input_status & 0x01) == 0) {                                         // no left horizontal shot
            if (sys_input_status & 0x04)    man_game_player_vshot(e);                 // vertical shot
            if (sys_input_status & 0x02)    man_game_player_hshot(e, SHOT_RIGHT);     // right horizontal shot
            return;
        }

        if ((sys_input_status & 0x02) == 0) {                                         // no right horizontal shot
            if (sys_input_status & 0x04)    man_game_player_vshot(e);                 // vertical shot
            if (sys_input_status & 0x01)    man_game_player_hshot(e, SHOT_LEFT);      // left horizontal shot

            return;
        }      

        if ((sys_input_status & 0x04) == 0) {                                         // no vertical shot
            if (sys_input_status & 0x01)    man_game_player_hshot(e, SHOT_LEFT);      // left horizontal shot
            if (sys_input_status & 0x02)    man_game_player_hshot(e, SHOT_RIGHT);     // right horizontal shot
            return;
        }

    }

}



////////////////////////////////////////////////////////////////////////////////
// UPDATE ONE ENTITY
//      Updates physics of a given entity
// Inputs:
//      e: Pointer to the entity to be updated
//
// Since we use the signature e_cmp_movable in the call from forall_matching 
// in physics_update, we know that all the entities passed as parameters have 
// e_cmp_movable = 1
void sys_physics_update_one_entity(Entity_t* e) {

    // Freeze entities in its place for game over animation
    if (man_livesboard_check() == 0) return;                              

    // If entity has input, check keyboard
    if (e->cmps & e_cmp_input) {

        switch (m_abduction_light) {
            case 0: // No abduction light
                    e->cmps |= e_cmp_render;
                    sys_input_update();
                    sys_physics_update_velocity(e); 
                    sys_physics_update_shots_status(e);
                    break;
            case 1: // Abduction light going down
                    //e->cmps &= ~e_cmp_render;             // It does not show the character getting abducted
                    return;
            
            case 2: // Abduction light going up
                    e->cmps &= ~e_cmp_render;
                    return;
        }

    }
    
    e->x += e->vx;
    e->y += e->vy;

}


//==============================================================================
//==============================================================================
// PUBLIC FUNCTIONS
//==============================================================================
//==============================================================================

////////////////////////////////////////////////////////////////////////////////
// UPDATE
//      Updates the physics system
//
// Control inversion:
// The system call the entity manager that loops through all the entities
// calling the physics update for one entity.
// (we provide a pointer to the function to update the physics of one entity)
//
// The system just moves the entities that can be moved (signature movable bit = 1)
// Matching the signature we avoid calling the function for entities that can't be moved.
void sys_physics_update() {
    man_entity_forall_matching( sys_physics_update_one_entity, e_cmp_movable );
}