//-----------------------------LICENSE NOTICE------------------------------------
//
////////////////////////////////////////////////////////////////////////////////
///////////////////////////  SHOOT AND RUN /////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
//
//  Code & Graphics Copyright (C) 2020 Rafa Gutierrez (rafagutierrez@qq.com)
//
//  This file is part of Shoot and run: an Amstrad CPC Game made with CPCTelera
//  for CPCRetroDev 2020
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.


#include "render.h"
#include <sprites/palette_0.h>
#include <man/game.h>
#include <man/level.h>
#include <man/entity_templates.h>


//==============================================================================
//==============================================================================
// PRIVATE VARIABLES
//==============================================================================
//==============================================================================

u8 background;

//==============================================================================
//==============================================================================
// PRIVATE FUNCTIONS
//==============================================================================
//==============================================================================

////////////////////////////////////////////////////////////////////////////////
// SETS THE BACKGROUND COLOR
//
void sys_render_setBackground(u8 c) {
   background = c;
}


////////////////////////////////////////////////////////////////////////////////
// RENDER UNDER GROUND
//

void sys_render_underground() {

      u8 c = cpct_px2byteM0(15, 15);
      u8 *pvmem = cpct_getScreenPtr (CPCT_VMEM_START, 0, GROUND_Y+2);               
      
      cpct_drawSolidBox(pvmem, c, 40, SCR_LAST_LINE-GROUND_Y-1);             
      pvmem += 40;                                          // CPC screen width = 80 bytes, so 2nd half is 40 bytes to the right
      cpct_drawSolidBox(pvmem, c, 40, SCR_LAST_LINE-GROUND_Y-1);

}

////////////////////////////////////////////////////////////////////////////////
// RENDER BACKGROUND
//
void sys_render_background() {

      u8 c = cpct_px2byteM0(1, 1);
      u8 *pvmem = cpct_getScreenPtr (CPCT_VMEM_START, 0, 112);               
      
      cpct_drawSolidBox(pvmem, c, 40, 65);             
      pvmem += 40;                                          // CPC screen width = 80 bytes, so 2nd half is 40 bytes to the right
      cpct_drawSolidBox(pvmem, c, 40, 65);

}


////////////////////////////////////////////////////////////////////////////////
// RENDER THE GROUND LINE
//
void sys_render_ground() {

   u8 c = cpct_px2byteM0(10, 10);
   u8 *pvmem = cpct_getScreenPtr (CPCT_VMEM_START, 0, GROUND_Y);     
 
   cpct_drawSolidBox(pvmem, c, 40, 2);
   pvmem += 40;                                                // CPC screen width = 80 bytes, so 2nd half is 40 bytes to the right
   cpct_drawSolidBox(pvmem, c, 40, 2);

}


////////////////////////////////////////////////////////////////////////////////
// RENDER ONE ENTITY
//      Renders a given entity to video memory
//      If entity is dead  -> erase it  (drawSolidBox)
//      If entity is alive -> render it (drawSprite)
// Inputs:
//      e: Pointer to the entity to be rendered
//
// TODO: local variables and save current pvmen in entity so when we need to 
// delete entity in previous position we do not need to calculate it again
//
// Have some kind of counter so we do not render in every frame 
//
void sys_render_one_entity ( Entity_t* e ) {
   u8* pvmem = cpct_getScreenPtr (CPCT_VMEM_START, e->x, e->y);     

   if (e->type & e_type_dead)                            // dead entity: erase from screen
      cpct_drawSolidBox(pvmem, 0, e->w, e->h);              
   else                                                  // alive entity: draw its sprite
      if (e->type & e_type_enemy_shot) {
         
         // draw shot in current position
         cpct_drawSprite(e->sprite, pvmem, e->w, e->h);

         // restore background from previous position
 
      }
      else 
         cpct_drawSprite(e->sprite, pvmem, e->w, e->h);
   
}


void sys_render_one_entity_bak ( Entity_t* e ) {
   u8* pvmem = cpct_getScreenPtr (CPCT_VMEM_START, e->x, e->y);     

   if (e->type & e_type_dead)                            // dead entity: erase from screen
      cpct_drawSolidBox(pvmem, 0, e->w, e->h);              
   else                                                  // alive entity: draw its sprite
      cpct_drawSprite(e->sprite, pvmem, e->w, e->h);
   
}



//==============================================================================
//==============================================================================
// PUBLIC FUNCTIONS
//==============================================================================
//==============================================================================

////////////////////////////////////////////////////////////////////////////////
// INIT
//      Initializes rendering
//
void sys_render_init() {

   cpct_clearScreen(0);
   cpct_setVideoMode(0);                    // In mode 0 each byte = 2 pixels
   cpct_setPalette(palette_0, 16);
   cpct_setPALColour(15, HW_BLACK);         // the pink color (15) we change to black but different than the background
  
   background = HW_BLACK;                   // Sets background colour
   cpct_setBorder(HW_BLACK);                // Sets border colour

 //  sys_render_background();
   sys_render_underground();                // Renders the side under the ground with a 2nd black
   sys_render_ground();                     // Renders ground
    
}

////////////////////////////////////////////////////////////////////////////////
// UPDATE
//      Updates the render system
//
// Control inversion:
// The system call the entity manager that loops through all the entities
// calling the render update for one entity.
// (we provide a pointer to the function to update the physics of one entity)
//
// The system just render the entities that can be rendered (signature render bit = 1)
// Matching the signature we avoid calling the function for entities that can't be rendered.
void sys_render_update() {
  
   cpct_setPALColour (0, background);

   sys_render_ground();                     

   man_entity_forall_matching( sys_render_one_entity, e_cmp_render );
    
}



////////////////////////////////////////////////////////////////////////////////
// Instead of using c = cpct_px2byteM0(13, 13) we can calculate the value of c as:
//
// Palette color 13 in both pixels
//    13 = 0X0D = 1101
//               __BYTE__
//    _LP_ _RP_  0 2 1 3   (LP=Left pixel, RP=Right pixel)
//    3210 DCBA   A C B D  (Bit order 0A2C1B3D)
//    1101 1101  11110011  (=0xF3)
